"""API client functions."""

import time

import log
import requests

from . import API
from .cache import Cache

cache = Cache()


def put(slug: str, data):
    url = f"{API}/{slug}"
    log.info("PUT %s: %s", url, data)

    response = cache.get((url, data))
    if response is None:
        for delay in [1, 3, 5]:
            response = requests.put(url, data=data, timeout=30)
            if response.status_code == 500:
                time.sleep(delay)
                continue
            break
        cache.set((url, data), response)

    log.info("Response: %s", response)

    return response


def delete(slug: str):
    url = f"{API}/{slug}"
    log.info("DELETE %s", url)

    for delay in [1, 3, 5]:
        response = requests.delete(url, timeout=30)
        if response.status_code == 500:
            time.sleep(delay)
            continue
        break

    log.info("Response: %s", response)

    cache.clear()

    return response



"""Utilities to detect the environment this program is running on."""

import os
from pathlib import Path

CONTINUOUS_INTEGRATION = [
    # General
    "CI",
    "CONTINUOUS_INTEGRATION",
    "DISABLE_COVERAGE",
    # Travis CI
    "TRAVIS",
    # Appveyor
    "APPVEYOR",
    # CircleCI
    "CIRCLECI",
    # Drone
    "DRONE",
]


def ci():
    return any(name in CONTINUOUS_INTEGRATION for name in os.environ)


def poetry():
    return "POETRY_ACTIVE" in os.environ or Path("poetry.lock").exists()



"""Plugins to extract coverage data from various formats."""

import os
import time
import webbrowser
from abc import ABC, abstractmethod

import coverage
import log

from .cache import Cache

cache = Cache()


class BasePlugin(ABC):
    """Base class for coverage plugins."""

    @abstractmethod
    def matches(self, cwd):
        """Determine if the current directory contains coverage data.

        :return bool: Indicates that the current directory should be processed.

        """
        raise NotImplementedError

    @abstractmethod
    def get_coverage(self, cwd):
        """Extract the coverage data from the current directory.

        :return float: Percentage of lines covered.

        """
        raise NotImplementedError

    @abstractmethod
    def get_report(self, cwd):
        """Get the path to the coverage report.

        :return str: Path to coverage report or `None` if not available.

        """
        raise NotImplementedError


def get_coverage(cwd=None, *, always=False) -> float:
    """Extract the current coverage data."""
    cwd = cwd or os.getcwd()

    plugin = _find_plugin(cwd, allow_missing=always)
    if plugin is None:
        return 0.0

    percentage = plugin.get_coverage(cwd)
    return round(percentage, 1)


def get_report(cwd=None):
    """Get the full path to the coverage report.."""
    cwd = cwd or os.getcwd()

    plugin = _find_plugin(cwd, allow_missing=True)

    return plugin.get_report(cwd) if plugin else None


def launch_report(cwd=None, *, always=False):
    """Open the generated coverage report in a web browser."""
    cwd = cwd or os.getcwd()

    plugin = _find_plugin(cwd, allow_missing=True)

    if plugin:
        report = plugin.get_report(cwd)

        if report and (not _launched_recently(report) or always):
            log.info("Launching report: %s", report)
            webbrowser.open("file://" + report, new=2, autoraise=True)
        elif always:
            log.error("No coverage report found")


def _find_plugin(cwd, allow_missing=False):
    """Find an return a matching coverage plugin."""
    for cls in BasePlugin.__subclasses__():  # pylint: disable=no-member
        plugin = cls()  # type: ignore
        if plugin.matches(cwd):
            return plugin

    msg = f"No coverage data found: {cwd}"
    log.info(msg)

    if allow_missing:
        return None

    raise RuntimeError(msg)


def _launched_recently(path):
    now = time.time()
    then = cache.get(path, default=0)
    elapsed = round(now - then, 1)
    log.info("Report last launched %s seconds ago", elapsed)
    cache.set(path, now)
    return elapsed < 60 * 60  # 1 hour


class CoveragePy(BasePlugin):
    """Coverage extractor for the coverage.py format."""

    def matches(self, cwd):
        return any((".coverage" in os.listdir(cwd), ".coveragerc" in os.listdir(cwd)))

    def get_coverage(self, cwd):
        os.chdir(cwd)

        cov = coverage.Coverage()
        cov.load()

        with open(os.devnull, "w", encoding="utf-8") as ignore:
            total = cov.report(file=ignore)

        return total

    def get_report(self, cwd):
        for root, _, _ in os.walk(cwd):
            path = os.path.join(root, "htmlcov", "index.html")

            if os.path.isfile(path):
                log.info("Found coverage report: %s", path)
                return path

        log.info("No coverage report found: %s", cwd)
        return None



"""Update project metrics on The Coverage Space.

Usage:
  coveragespace [--verbose]
  coveragespace update <metric> [<value>] [--verbose] [--exit-code]
  coveragespace reset [--verbose]
  coveragespace view [--verbose]
  coveragespace (-h | --help)
  coveragespace (-V | --version)

Options:
  -h --help         Show this help screen.
  -V --version      Show the program version.
  -v --verbose      Always display the coverage metrics.
  -x --exit-code    Return non-zero exit code on failures.

"""

import json
import sys
from shutil import get_terminal_size

import colorama
import log
from docopt import docopt

from . import VERSION, api, coverage, environments, repository


def main():
    """Parse command-line arguments, configure logging, and run the program."""
    colorama.init(autoreset=True)
    arguments = docopt(__doc__, version=VERSION)
    log.init(level=log.DEBUG if arguments["--verbose"] else log.WARNING)

    if environments.ci():
        log.info("Coverage check skipped when running on CI service")
        sys.exit()

    if arguments["view"]:
        success = view()
    else:
        slug = repository.get_slug()
        if arguments["update"]:
            success = call(
                slug,
                arguments["<metric>"],
                arguments["<value>"],
                verbose=arguments["--verbose"],
                hardfail=arguments["--exit-code"],
            )
        elif arguments["reset"]:
            success = call(slug, reset=True)
        else:
            success = call(slug, launch=True)

    if not success and arguments["--exit-code"]:
        sys.exit(1)


def call(
    slug: str,
    metric: str = "overall",
    value: float = 0,
    *,
    reset: bool = False,
    launch: bool = False,
    verbose: bool = False,
    hardfail: bool = False,
):
    """Call the API and display errors."""
    if reset:
        response = api.delete(slug)
    else:
        data = {metric: value or coverage.get_coverage(always=launch)}
        response = api.put(slug, data)

    if response.status_code == 200:
        if verbose or launch:
            data = response.json()
            report = coverage.get_report()
            if report:
                data["report"] = report
            display("coverage updated", data, colorama.Fore.GREEN)
        if launch:
            coverage.launch_report(always=True)
        return True

    if response.status_code == 202:
        data = response.json()
        report = coverage.get_report()
        if report:
            data["report"] = report
        display("coverage reset", data, colorama.Fore.BLUE)
        return True

    if response.status_code == 422:
        color = colorama.Fore.RED if hardfail else colorama.Fore.YELLOW
        data = response.json()
        prefix = "poetry run " if environments.poetry() else ""
        report = coverage.get_report()
        if report:
            data["report"] = report
        data["help"] = f"To reset metrics, run: ^{prefix}coveragespace reset$"  # type: ignore
        display("coverage decreased", data, color)
        coverage.launch_report(always=launch)
        return False

    try:
        data = response.json()
        display("coverage unknown", data, colorama.Fore.RED)
    except (TypeError, ValueError) as exc:
        data = response.data.decode("utf-8")
        log.error("%s\n\nwhen decoding response:\n\n%s\n", exc, data)
    return False


def display(title, data, color=""):
    """Write colored text to the console."""
    width, _ = get_terminal_size()
    header = color + f"{' ' + title + ' ':=^{width}}"
    header = header.replace(
        title, colorama.Style.BRIGHT + title + colorama.Style.NORMAL
    )
    body = json.dumps(data, indent=4)
    body = body.replace("^", colorama.Fore.WHITE + colorama.Style.BRIGHT)
    body = body.replace("$", colorama.Style.RESET_ALL)
    footer = color + "=" * width
    print(header)
    print(body)
    print(footer)


def view():
    """View the local coverage report."""
    coverage.launch_report(always=True)
    return True



# pylint: disable=unused-variable,unused-argument,expression-not-assigned,singleton-comparison,redefined-outer-name

import os
import time
from unittest.mock import Mock, patch

import pytest

from ..coverage import _launched_recently, cache, get_coverage


class MockCoverage(Mock):
    @staticmethod
    def report(*args, **kwargs):
        return 42.456


@pytest.fixture
def coveragepy_data(tmpdir):
    cwd = tmpdir.chdir()
    with open("foobar.py", "w", encoding="utf-8") as stream:
        pass
    with open(".coverage", "w", encoding="utf-8") as stream:
        stream.write(
            """
            !coverage.py: This is a private format, don\'t read it directly!
            {"arcs":{"foobar.py": [[-1, 2]]}}
            """
        )


@pytest.fixture
def coveragepy_data_custom(tmpdir):
    cwd = tmpdir.chdir()
    with open("foobar.py", "w", encoding="utf-8") as stream:
        pass
    with open(".coveragerc", "w", encoding="utf-8") as stream:
        stream.write(
            """
            [run]
            data_file = .cache/coverage
            """
        )
    os.makedirs(".cache")
    with open(".cache/coverage", "w", encoding="utf-8") as stream:
        stream.write(
            """
            !coverage.py: This is a private format, don\'t read it directly!
            {"arcs":{"foobar.py": [[-1, 3]]}}
            """
        )


def describe_get_coverage():
    @patch("coverage.Coverage", MockCoverage)
    def it_supports_coveragepy(expect, coveragepy_data):
        expect(get_coverage()) == 42.5

    @patch("coverage.Coverage", MockCoverage)
    def it_supports_coveragepy_with_custom_location(expect, coveragepy_data_custom):
        expect(get_coverage()) == 42.5


def describe_launched_recently():
    def when_never_launched(expect):
        cache.set("mock/path", 0)
        expect(_launched_recently("mock/path")) == False

    def when_just_launched(expect):
        cache.set("mock/path", time.time())
        expect(_launched_recently("mock/path")) == True

    def when_launched_59_minutes_ago(expect):
        cache.set("mock/path", time.time() - 60 * 59)
        expect(_launched_recently("mock/path")) == True

    def when_launched_61_minutes_ago(expect):
        cache.set("mock/path", time.time() - 60 * 61)
        expect(_launched_recently("mock/path")) == False
